"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const sql = require("pg-sql2");
const util_1 = require("util");
// @ts-ignore
const package_json_1 = require("../../package.json");
const pgField_1 = require("./pgField");
const queryFromResolveDataFactory_1 = require("../queryFromResolveDataFactory");
const addStartEndCursor_1 = require("./addStartEndCursor");
const omit_1 = require("../omit");
const makeProcField_1 = require("./makeProcField");
const parseIdentifier_1 = require("../parseIdentifier");
const viaTemporaryTable_1 = require("./viaTemporaryTable");
const chalk_1 = require("chalk");
const pickBy = require("lodash/pickBy");
const PgLiveProvider_1 = require("../PgLiveProvider");
const pgPrepareAndRun_1 = require("../pgPrepareAndRun");
/**
 * Given a PgEntity and a tag name, fetches the tag with that name. If it
 * exists, asserts that it is a string (and not a boolean, or array). If it
 * doesn't exist, returns null.
 */
function stringTag(pgEntity, tagName) {
    const tagVal = pgEntity.tags[tagName] || null;
    if (typeof tagVal === "string") {
        return tagVal;
    }
    else if (!tagVal) {
        return null;
    }
    else {
        throw new Error(`Expected smart tag '${tagName}' on '${describePgEntity(pgEntity)}' to be a string, but instead received '${util_1.inspect(tagVal)}'`);
    }
}
exports.stringTag = stringTag;
function makePgBaseInflectors() {
    const inflectors = {
        // These helpers are passed GraphQL type names as strings
        conditionType(typeName) {
            return this.upperCamelCase(`${typeName}-condition`);
        },
        inputType(typeName) {
            return this.upperCamelCase(`${typeName}-input`);
        },
        rangeBoundType(typeName) {
            return this.upperCamelCase(`${typeName}-range-bound`);
        },
        rangeType(typeName) {
            return this.upperCamelCase(`${typeName}-range`);
        },
        patchType(typeName) {
            return this.upperCamelCase(`${typeName}-patch`);
        },
        baseInputType(typeName) {
            return this.upperCamelCase(`${typeName}-base-input`);
        },
        patchField(itemName) {
            return this.camelCase(`${itemName}-patch`);
        },
        orderByType(typeName) {
            return this.upperCamelCase(`${this.pluralize(typeName)}-order-by`);
        },
        edge(typeName) {
            return this.upperCamelCase(`${this.pluralize(typeName)}-edge`);
        },
        connection(typeName) {
            return this.upperCamelCase(`${this.pluralize(typeName)}-connection`);
        },
        // These helpers handle overrides via smart comments. They should only
        // be used in other inflectors, hence the underscore prefix.
        //
        // IMPORTANT: do NOT do case transforms here, because detail can be
        // lost, e.g.
        // `constantCase(camelCase('foo_1')) !== constantCase('foo_1')`
        _functionName(proc) {
            return this.coerceToGraphQLName(stringTag(proc, "name") || proc.name);
        },
        _typeName(type) {
            // 'type' introspection result
            return this.coerceToGraphQLName(stringTag(type, "name") || type.name);
        },
        _tableName(table) {
            return this.coerceToGraphQLName(stringTag(table, "name") || stringTag(table.type, "name") || table.name);
        },
        _singularizedTableName(table) {
            return this.singularize(this._tableName(table)).replace(/.(?:(?:[_-]i|I)nput|(?:[_-]p|P)atch)$/, "$&_record");
        },
        _columnName(attr, _options) {
            return this.coerceToGraphQLName(stringTag(attr, "name") || attr.name);
        },
        // From here down, functions are passed database introspection results
        enumType(type) {
            return this.upperCamelCase(this._typeName(type));
        },
        argument(name, index) {
            return this.coerceToGraphQLName(this.camelCase(name || `arg${index}`));
        },
        orderByEnum(columnName, ascending) {
            return this.constantCase(`${columnName}_${ascending ? "asc" : "desc"}`);
        },
        orderByColumnEnum(attr, ascending) {
            const columnName = this._columnName(attr, {
                skipRowId: true,
            });
            return this.orderByEnum(columnName, ascending);
        },
        orderByComputedColumnEnum(pseudoColumnName, proc, table, ascending) {
            const columnName = this.computedColumn(pseudoColumnName, proc, table);
            return this.orderByEnum(columnName, ascending);
        },
        domainType(type) {
            return this.upperCamelCase(this._typeName(type));
        },
        enumName(inValue) {
            let value = inValue;
            if (value === "") {
                return "_EMPTY_";
            }
            // Some enums use asterisks to signify wildcards - this might be for
            // the whole item, or prefixes/suffixes, or even in the middle.  This
            // is provided on a best efforts basis, if it doesn't suit your
            // purposes then please pass a custom inflector as mentioned below.
            value = value
                .replace(/\*/g, "_ASTERISK_")
                .replace(/^(_?)_+ASTERISK/, "$1ASTERISK")
                .replace(/ASTERISK_(_?)_*$/, "ASTERISK$1");
            // This is a best efforts replacement for common symbols that you
            // might find in enums. Generally we only support enums that are
            // alphanumeric, if these replacements don't work for you, you should
            // pass a custom inflector that replaces this `enumName` method
            // with one of your own chosing.
            value =
                {
                    // SQL comparison operators
                    ">": "GREATER_THAN",
                    ">=": "GREATER_THAN_OR_EQUAL",
                    "=": "EQUAL",
                    "!=": "NOT_EQUAL",
                    "<>": "DIFFERENT",
                    "<=": "LESS_THAN_OR_EQUAL",
                    "<": "LESS_THAN",
                    // PostgreSQL LIKE shortcuts
                    "~~": "LIKE",
                    "~~*": "ILIKE",
                    "!~~": "NOT_LIKE",
                    "!~~*": "NOT_ILIKE",
                    // '~' doesn't necessarily represent regexps, but the three
                    // operators following it likely do, so we'll use the word TILDE
                    // in all for consistency.
                    "~": "TILDE",
                    "~*": "TILDE_ASTERISK",
                    "!~": "NOT_TILDE",
                    "!~*": "NOT_TILDE_ASTERISK",
                    // A number of other symbols where we're not sure of their
                    // meaning.  We give them common generic names so that they're
                    // suitable for multiple purposes, e.g. favouring 'PLUS' over
                    // 'ADDITION' and 'DOT' over 'FULL_STOP'
                    "%": "PERCENT",
                    "+": "PLUS",
                    "-": "MINUS",
                    "/": "SLASH",
                    "\\": "BACKSLASH",
                    _: "UNDERSCORE",
                    "#": "POUND",
                    "£": "STERLING",
                    $: "DOLLAR",
                    "&": "AMPERSAND",
                    "@": "AT",
                    "'": "APOSTROPHE",
                    '"': "QUOTE",
                    "`": "BACKTICK",
                    ":": "COLON",
                    ";": "SEMICOLON",
                    "!": "EXCLAMATION_POINT",
                    "?": "QUESTION_MARK",
                    ",": "COMMA",
                    ".": "DOT",
                    "^": "CARET",
                    "|": "BAR",
                    "[": "OPEN_BRACKET",
                    "]": "CLOSE_BRACKET",
                    "(": "OPEN_PARENTHESIS",
                    ")": "CLOSE_PARENTHESIS",
                    "{": "OPEN_BRACE",
                    "}": "CLOSE_BRACE",
                }[value] || value;
            return value;
        },
        tableNode(table) {
            return this.camelCase(this._singularizedTableName(table));
        },
        tableFieldName(table) {
            return this.camelCase(this._singularizedTableName(table));
        },
        allRows(table) {
            return this.camelCase(`all-${this.pluralize(this._singularizedTableName(table))}`);
        },
        allRowsSimple(table) {
            return this.camelCase(`all-${this.pluralize(this._singularizedTableName(table))}-list`);
        },
        functionMutationName(proc) {
            return this.camelCase(this._functionName(proc));
        },
        functionMutationResultFieldName(proc, gqlType, plural = false, outputArgNames = []) {
            const resultFieldName = stringTag(proc, "resultFieldName");
            if (resultFieldName) {
                return resultFieldName;
            }
            let name;
            if (outputArgNames.length === 1 && outputArgNames[0] !== "") {
                name = this.camelCase(outputArgNames[0]);
            }
            else if (gqlType.name === "Int") {
                name = "integer";
            }
            else if (gqlType.name === "Float") {
                name = "float";
            }
            else if (gqlType.name === "Boolean") {
                name = "boolean";
            }
            else if (gqlType.name === "String") {
                name = "string";
            }
            else if (proc.returnTypeId === "2249") {
                // returns a record type
                name = "result";
            }
            else {
                name = this.camelCase(gqlType.name);
            }
            return plural ? this.pluralize(name) : name;
        },
        functionQueryName(proc) {
            return this.camelCase(this._functionName(proc));
        },
        functionQueryNameList(proc) {
            return this.camelCase(`${this._functionName(proc)}-list`);
        },
        functionPayloadType(proc) {
            return this.upperCamelCase(`${this._functionName(proc)}-payload`);
        },
        functionInputType(proc) {
            return this.upperCamelCase(`${this._functionName(proc)}-input`);
        },
        functionOutputFieldName(_proc, outputArgName, index) {
            return this.argument(outputArgName, index);
        },
        tableType(table) {
            return this.upperCamelCase(this._singularizedTableName(table));
        },
        column(attr) {
            return this.camelCase(this._columnName(attr));
        },
        computedColumn(pseudoColumnName, proc, _table) {
            return stringTag(proc, "fieldName") || this.camelCase(pseudoColumnName);
        },
        computedColumnList(pseudoColumnName, proc, _table) {
            return stringTag(proc, "fieldName")
                ? stringTag(proc, "fieldName") + "List"
                : this.camelCase(`${pseudoColumnName}-list`);
        },
        singleRelationByKeys(detailedKeys, table, _foreignTable, constraint) {
            const fieldName = stringTag(constraint, "fieldName");
            if (fieldName) {
                return fieldName;
            }
            return this.camelCase(`${this._singularizedTableName(table)}-by-${detailedKeys
                .map(key => this.column(key))
                .join("-and-")}`);
        },
        singleRelationByKeysBackwards(detailedKeys, table, _foreignTable, constraint) {
            const foreignSingleFieldName = stringTag(constraint, "foreignSingleFieldName");
            if (foreignSingleFieldName) {
                return foreignSingleFieldName;
            }
            const foreignFieldName = stringTag(constraint, "foreignFieldName");
            if (foreignFieldName) {
                return foreignFieldName;
            }
            return this.singleRelationByKeys(detailedKeys, table, _foreignTable, constraint);
        },
        manyRelationByKeys(detailedKeys, table, _foreignTable, constraint) {
            const foreignFieldName = stringTag(constraint, "foreignFieldName");
            if (foreignFieldName) {
                return foreignFieldName;
            }
            return this.camelCase(`${this.pluralize(this._singularizedTableName(table))}-by-${detailedKeys.map(key => this.column(key)).join("-and-")}`);
        },
        manyRelationByKeysSimple(detailedKeys, table, _foreignTable, constraint) {
            const foreignSimpleFieldName = stringTag(constraint, "foreignSimpleFieldName");
            if (foreignSimpleFieldName) {
                return foreignSimpleFieldName;
            }
            const foreignFieldName = stringTag(constraint, "foreignFieldName");
            if (foreignFieldName) {
                return foreignFieldName;
            }
            return this.camelCase(`${this.pluralize(this._singularizedTableName(table))}-by-${detailedKeys.map(key => this.column(key)).join("-and-")}-list`);
        },
        rowByUniqueKeys(detailedKeys, table, constraint) {
            const fieldName = stringTag(constraint, "fieldName");
            if (fieldName) {
                return fieldName;
            }
            return this.camelCase(`${this._singularizedTableName(table)}-by-${detailedKeys
                .map(key => this.column(key))
                .join("-and-")}`);
        },
        updateByKeys(detailedKeys, table, constraint) {
            const updateFieldName = stringTag(constraint, "updateFieldName");
            if (updateFieldName) {
                return updateFieldName;
            }
            return this.camelCase(`update-${this._singularizedTableName(table)}-by-${detailedKeys
                .map(key => this.column(key))
                .join("-and-")}`);
        },
        deleteByKeys(detailedKeys, table, constraint) {
            const deleteFieldName = stringTag(constraint, "deleteFieldName");
            if (deleteFieldName) {
                return deleteFieldName;
            }
            return this.camelCase(`delete-${this._singularizedTableName(table)}-by-${detailedKeys
                .map(key => this.column(key))
                .join("-and-")}`);
        },
        updateByKeysInputType(detailedKeys, table, constraint) {
            const updateFieldName = stringTag(constraint, "updateFieldName");
            if (updateFieldName) {
                return this.upperCamelCase(`${updateFieldName}-input`);
            }
            return this.upperCamelCase(`update-${this._singularizedTableName(table)}-by-${detailedKeys
                .map(key => this.column(key))
                .join("-and-")}-input`);
        },
        deleteByKeysInputType(detailedKeys, table, constraint) {
            const deleteFieldName = stringTag(constraint, "deleteFieldName");
            if (deleteFieldName) {
                return this.upperCamelCase(`${deleteFieldName}-input`);
            }
            return this.upperCamelCase(`delete-${this._singularizedTableName(table)}-by-${detailedKeys
                .map(key => this.column(key))
                .join("-and-")}-input`);
        },
        updateNode(table) {
            return this.camelCase(`update-${this._singularizedTableName(table)}`);
        },
        deleteNode(table) {
            return this.camelCase(`delete-${this._singularizedTableName(table)}`);
        },
        deletedNodeId(table) {
            return this.camelCase(`deleted-${this.singularize(table.name)}-id`);
        },
        updateNodeInputType(table) {
            return this.upperCamelCase(`update-${this._singularizedTableName(table)}-input`);
        },
        deleteNodeInputType(table) {
            return this.upperCamelCase(`delete-${this._singularizedTableName(table)}-input`);
        },
        edgeField(table) {
            return this.camelCase(`${this._singularizedTableName(table)}-edge`);
        },
        recordFunctionReturnType(proc) {
            return (stringTag(proc, "resultTypeName") ||
                this.upperCamelCase(`${this._functionName(proc)}-record`));
        },
        recordFunctionConnection(proc) {
            return this.upperCamelCase(`${this._functionName(proc)}-connection`);
        },
        recordFunctionEdge(proc) {
            return this.upperCamelCase(`${this.singularize(this._functionName(proc))}-edge`);
        },
        scalarFunctionConnection(proc) {
            return this.upperCamelCase(`${this._functionName(proc)}-connection`);
        },
        scalarFunctionEdge(proc) {
            return this.upperCamelCase(`${this.singularize(this._functionName(proc))}-edge`);
        },
        createField(table) {
            return this.camelCase(`create-${this._singularizedTableName(table)}`);
        },
        createInputType(table) {
            return this.upperCamelCase(`create-${this._singularizedTableName(table)}-input`);
        },
        createPayloadType(table) {
            return this.upperCamelCase(`create-${this._singularizedTableName(table)}-payload`);
        },
        updatePayloadType(table) {
            return this.upperCamelCase(`update-${this._singularizedTableName(table)}-payload`);
        },
        deletePayloadType(table) {
            return this.upperCamelCase(`delete-${this._singularizedTableName(table)}-payload`);
        },
    };
    return preventEmptyResult(inflectors);
}
const defaultPgColumnFilter = (_attr, _build, _context) => true;
function identity(val) {
    return val;
}
function preventEmptyResult(obj) {
    return Object.keys(obj).reduce((memo, key) => {
        const fn = obj[key];
        memo[key] = function (...args) {
            const result = fn.apply(this, args);
            if (typeof result !== "string" || result.length === 0) {
                const stringifiedArgs = require("util").inspect(args);
                throw new Error(`Inflector for '${key}' returned '${String(result)}'; expected non-empty string\n` +
                    `See: https://github.com/graphile/graphile-build/blob/master/packages/graphile-build-pg/src/plugins/PgBasicsPlugin.js\n` +
                    `Arguments passed to ${key}:\n${stringifiedArgs}`);
            }
            return result;
        };
        return memo;
    }, {});
}
exports.preventEmptyResult = preventEmptyResult;
const omitWithRBACChecks = (omit) => (entity, permission) => {
    const ORDINARY_TABLE = "r";
    const VIEW = "v";
    const MATERIALIZED_VIEW = "m";
    const isTableLike = (entity) => entity &&
        entity.kind === "class" &&
        (entity.classKind === ORDINARY_TABLE ||
            entity.classKind === VIEW ||
            entity.classKind === MATERIALIZED_VIEW);
    if (entity.kind === "procedure") {
        if (permission === omit_1.EXECUTE && !entity.aclExecutable) {
            return true;
        }
    }
    else if (entity.kind === "class" && isTableLike(entity)) {
        const tableEntity = entity;
        if ((permission === omit_1.READ || permission === omit_1.ALL || permission === omit_1.MANY) &&
            !tableEntity.aclSelectable &&
            !tableEntity.attributes.some(attr => attr.aclSelectable)) {
            return true;
        }
        else if (permission === omit_1.CREATE &&
            !tableEntity.aclInsertable &&
            !tableEntity.attributes.some(attr => attr.aclInsertable)) {
            return true;
        }
        else if (permission === omit_1.UPDATE &&
            !tableEntity.aclUpdatable &&
            !tableEntity.attributes.some(attr => attr.aclUpdatable)) {
            return true;
        }
        else if (permission === omit_1.DELETE && !tableEntity.aclDeletable) {
            return true;
        }
    }
    else if (entity.kind === "attribute" && isTableLike(entity.class)) {
        const attributeEntity = entity;
        const klass = attributeEntity.class;
        // Have we got *any* permissions on the table?
        if (klass.aclSelectable ||
            klass.attributes.some(attr => attr.aclSelectable)) {
            // Yes; this is a regular table; omit if RBAC permissions tell us to.
            if ((permission === omit_1.READ ||
                permission === omit_1.FILTER ||
                permission === omit_1.ORDER) &&
                !attributeEntity.aclSelectable) {
                return true;
            }
            else if (permission === omit_1.CREATE && !attributeEntity.aclInsertable) {
                return true;
            }
            else if (permission === omit_1.UPDATE && !attributeEntity.aclUpdatable) {
                return true;
            }
        }
        else {
            // No permissions on the table at all, so normal connections will skip
            // over it. Thus we must be being exposed via a security definer function
            // or similar, so we should expose all fields except those that are
            // explicitly @omit-ed.
        }
    }
    return omit(entity, permission);
};
const omitUnindexed = (omit, hideIndexWarnings) => (entity, permission) => {
    if (entity.kind === "attribute" &&
        !entity.isIndexed &&
        (permission === "filter" || permission === "order")) {
        return true;
    }
    if (entity.kind === "constraint" &&
        entity.type === "f" &&
        !entity.isIndexed &&
        permission === "read") {
        const klass = entity.class;
        if (klass) {
            const shouldOutputWarning = !entity["_omitUnindexedReadWarningGiven"] && !hideIndexWarnings;
            if (shouldOutputWarning) {
                entity["_omitUnindexedReadWarningGiven"] = true;
                // eslint-disable-next-line no-console
                console.log("%s", `Disabled 'read' permission for ${describePgEntity(entity)} because it isn't indexed. For more information see https://graphile.org/postgraphile/best-practices/ To fix, perform\n\n  CREATE INDEX ON ${`"${klass.namespaceName}"."${klass.name}"`}("${entity.keyAttributes
                    .map(a => a.name)
                    .join('", "')}");`);
            }
        }
        return true;
    }
    return omit(entity, permission);
};
function describePgEntity(entity, includeAlias = true) {
    const getAlias = !includeAlias
        ? () => ""
        : () => {
            const tags = pickBy(entity.tags, (_value, key) => key === "name" || key.endsWith("Name"));
            if (Object.keys(tags).length) {
                return ` (with smart comments: ${chalk_1.default.bold(Object.keys(tags)
                    .map(t => `@${t} ${tags[t]}`)
                    .join(" | "))})`;
            }
            return "";
        };
    try {
        if (entity.kind === "constraint") {
            return `constraint ${chalk_1.default.bold(`"${entity.name}"`)} on ${describePgEntity(entity.class, false)}${getAlias()}`;
        }
        else if (entity.kind === "class") {
            // see pg_class.relkind https://www.postgresql.org/docs/10/static/catalog-pg-class.html
            const kind = {
                c: "composite type",
                f: "foreign table",
                p: "partitioned table",
                r: "table",
                v: "view",
                m: "materialized view",
            }[entity.classKind] || "table-like";
            return `${kind} ${chalk_1.default.bold(`"${entity.namespaceName}"."${entity.name}"`)}${getAlias()}`;
        }
        else if (entity.kind === "procedure") {
            return `function ${chalk_1.default.bold(`"${entity.namespaceName}"."${entity.name}"(...args...)`)}${getAlias()}`;
        }
        else if (entity.kind === "attribute") {
            return `column ${chalk_1.default.bold(`"${entity.name}"`)} on ${describePgEntity(entity.class, false)}${getAlias()}`;
        }
    }
    catch (e) {
        // eslint-disable-next-line no-console
        console.error("Error occurred while attempting to debug entity:", entity);
        // eslint-disable-next-line no-console
        console.error(e);
    }
    return `entity of kind '${entity.kind}' with ${"id" in entity && typeof entity.id === "string" ? `oid '${entity.id}'` : ""}`;
}
exports.describePgEntity = describePgEntity;
function sqlCommentByAddingTags(entity, tagsToAdd) {
    // NOTE: this function is NOT intended to be SQL safe; it's for
    // displaying in error messages. Nonetheless if you find issues with
    // SQL compatibility, please send a PR or issue.
    // Ref: https://www.postgresql.org/docs/current/static/sql-syntax-lexical.html#SQL-BACKSLASH-TABLE
    const escape = (str) => str.replace(/['\\\b\f\n\r\t]/g, chr => ({
        "\b": "\\b",
        "\f": "\\f",
        "\n": "\\n",
        "\r": "\\r",
        "\t": "\\t",
    }[chr] || "\\" + chr));
    // tagsToAdd is here twice to ensure that the keys in tagsToAdd come first, but that they also "win" any conflicts.
    const tags = Object.assign(Object.assign(Object.assign({}, tagsToAdd), entity.tags), tagsToAdd);
    const description = entity.description;
    const tagsSql = Object.keys(tags)
        .reduce((memo, tag) => {
        const tagValue = tags[tag];
        const valueArray = Array.isArray(tagValue) ? tagValue : [tagValue];
        const highlightOrNot = tag in tagsToAdd ? chalk_1.default.bold.green : identity;
        valueArray.forEach(value => {
            memo.push(highlightOrNot(`@${escape(escape(tag))}${value === true ? "" : " " + escape(escape(value))}`));
        });
        return memo;
    }, [])
        .join("\\n");
    const commentValue = `E'${tagsSql}${description ? "\\n" + escape(description) : ""}'`;
    let sqlThing;
    if (entity.kind === "class") {
        const identifier = `"${entity.namespaceName}"."${entity.name}"`;
        if (entity.classKind === "r") {
            sqlThing = `TABLE ${identifier}`;
        }
        else if (entity.classKind === "v") {
            sqlThing = `VIEW ${identifier}`;
        }
        else if (entity.classKind === "m") {
            sqlThing = `MATERIALIZED VIEW ${identifier}`;
        }
        else if (entity.classKind === "c") {
            sqlThing = `TYPE ${identifier}`;
        }
        else {
            sqlThing = `PLEASE_SEND_A_PULL_REQUEST_TO_FIX_THIS ${identifier}`;
        }
    }
    else if (entity.kind === "attribute") {
        sqlThing = `COLUMN "${entity.class.namespaceName}"."${entity.class.name}"."${entity.name}"`;
    }
    else if (entity.kind === "procedure") {
        sqlThing = `FUNCTION "${entity.namespaceName}"."${entity.name}"(...arg types go here...)`;
    }
    else if (entity.kind === "constraint") {
        // TODO: TEST!
        sqlThing = `CONSTRAINT "${entity.name}" ON "${entity.class.namespaceName}"."${entity.class.name}"`;
    }
    else {
        sqlThing = `UNKNOWN_ENTITY_PLEASE_SEND_A_PULL_REQUEST`;
    }
    return `COMMENT ON ${sqlThing} IS ${commentValue};`;
}
exports.default = (function PgBasicsPlugin(builder, { pgStrictFunctions = false, pgColumnFilter = defaultPgColumnFilter, pgIgnoreRBAC = false, pgIgnoreIndexes = true, // TODO:v5: change this to false
pgHideIndexWarnings = false, pgLegacyJsonUuid = false, }) {
    let pgOmit = omit_1.default;
    if (!pgIgnoreRBAC) {
        pgOmit = omitWithRBACChecks(pgOmit);
    }
    if (!pgIgnoreIndexes) {
        pgOmit = omitUnindexed(pgOmit, pgHideIndexWarnings);
    }
    builder.hook("build", build => {
        build.versions["graphile-build-pg"] = package_json_1.version;
        build.liveCoordinator.registerProvider(new PgLiveProvider_1.default());
        const buildExtensions = {
            graphileBuildPgVersion: package_json_1.version,
            pgSql: sql,
            pgStrictFunctions,
            pgColumnFilter,
            // TODO:v5: remove this workaround
            // BEWARE: this may be overridden in PgIntrospectionPlugin for PG < 9.5
            pgQueryFromResolveData: queryFromResolveDataFactory_1.default(),
            pgAddStartEndCursor: addStartEndCursor_1.default,
            pgOmit,
            pgMakeProcField: makeProcField_1.default,
            pgParseIdentifier: parseIdentifier_1.default,
            pgViaTemporaryTable: viaTemporaryTable_1.default,
            describePgEntity,
            pgField: pgField_1.default,
            sqlCommentByAddingTags,
            pgPrepareAndRun: pgPrepareAndRun_1.default,
        };
        return build.extend(build, buildExtensions, "Adding the graphile-build-pg basics");
    }, ["PgBasics"]);
    builder.hook("inflection", (inflection, build) => {
        // TODO:v5: move this to postgraphile-core
        const oldBuiltin = inflection.builtin;
        inflection.builtin = function (name) {
            if (pgLegacyJsonUuid && name === "JSON")
                return "Json";
            if (pgLegacyJsonUuid && name === "UUID")
                return "Uuid";
            return oldBuiltin.call(this, name);
        };
        return build.extend(inflection, makePgBaseInflectors(), "Default inflectors from PgBasicsPlugin. You can override these with `makeAddInflectorsPlugin(..., true)`.");
    }, ["PgBasics"]);
});
//# sourceMappingURL=PgBasicsPlugin.js.map