"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const withPgClient_1 = require("../withPgClient");
const utils_1 = require("../utils");
const fs_1 = require("fs");
const debug_1 = require("debug");
const chalk_1 = require("chalk");
const throttle = require("lodash/throttle");
const flatMap = require("lodash/flatMap");
const introspectionQuery_1 = require("./introspectionQuery");
const PgBasicsPlugin_1 = require("./PgBasicsPlugin");
// @ts-ignore
const package_json_1 = require("../../package.json");
const queryFromResolveDataFactory_1 = require("../queryFromResolveDataFactory");
const debug = debug_1.default("graphile-build-pg");
const WATCH_FIXTURES_PATH = `${__dirname}/../../res/watch-fixtures.sql`;
// Ref: https://github.com/graphile/postgraphile/tree/master/src/postgres/introspection/object
var PgEntityKind;
(function (PgEntityKind) {
    PgEntityKind["NAMESPACE"] = "namespace";
    PgEntityKind["PROCEDURE"] = "procedure";
    PgEntityKind["CLASS"] = "class";
    PgEntityKind["TYPE"] = "type";
    PgEntityKind["ATTRIBUTE"] = "attribute";
    PgEntityKind["CONSTRAINT"] = "constraint";
    PgEntityKind["EXTENSION"] = "extension";
    PgEntityKind["INDEX"] = "index";
})(PgEntityKind = exports.PgEntityKind || (exports.PgEntityKind = {}));
function readFile(filename, encoding) {
    return new Promise((resolve, reject) => {
        fs_1.readFile(filename, encoding, (err, res) => {
            if (err)
                reject(err);
            else
                resolve(res);
        });
    });
}
const removeQuotes = (str) => {
    const trimmed = str.trim();
    if (trimmed[0] === '"') {
        if (trimmed[trimmed.length - 1] !== '"') {
            throw new Error(`We failed to parse a quoted identifier '${str}'. Please avoid putting quotes or commas in smart comment identifiers (or file a PR to fix the parser).`);
        }
        return trimmed.substr(1, trimmed.length - 2);
    }
    else {
        // PostgreSQL lower-cases unquoted columns, so we should too.
        return trimmed.toLowerCase();
    }
};
const parseSqlColumnArray = (str) => {
    if (!str) {
        throw new Error(`Cannot parse '${str}'`);
    }
    const parts = str.split(",");
    return parts.map(removeQuotes);
};
const parseSqlColumnString = (str) => {
    if (!str) {
        throw new Error(`Cannot parse '${str}'`);
    }
    return removeQuotes(str);
};
function parseConstraintSpec(rawSpec) {
    const [spec, ...tagComponents] = rawSpec.split(/\|/);
    const parsed = utils_1.parseTags(tagComponents.join("\n"));
    return {
        spec,
        tags: parsed.tags,
        description: parsed.text,
    };
}
function smartCommentConstraints(introspectionResults) {
    const attributesByNames = (tbl, cols, debugStr) => {
        const attributes = introspectionResults.attribute
            .filter(a => a.classId === tbl.id)
            .sort((a, b) => a.num - b.num);
        if (!cols) {
            const pk = introspectionResults.constraint.find(c => c.classId == tbl.id && c.type === "p");
            if (pk) {
                return pk.keyAttributeNums.map(n => {
                    const attr = attributes.find(a => a.num === n);
                    if (!attr) {
                        throw new Error(`Could not find attribute '${n}' in '${PgBasicsPlugin_1.describePgEntity(pk)}'`);
                    }
                    return attr;
                });
            }
            else {
                throw new Error(`No columns specified for '${tbl.namespaceName}.${tbl.name}' (oid: ${tbl.id}) and no PK found (${debugStr}).`);
            }
        }
        return cols.map(colName => {
            const attr = attributes.find(a => a.name === colName);
            if (!attr) {
                throw new Error(`Could not find attribute '${colName}' in '${tbl.namespaceName}.${tbl.name}'`);
            }
            return attr;
        });
    };
    // First: primary keys
    introspectionResults.class.forEach(klass => {
        const namespace = introspectionResults.namespace.find(n => n.id === klass.namespaceId);
        if (!namespace) {
            return;
        }
        if (klass.tags.primaryKey) {
            if (typeof klass.tags.primaryKey !== "string") {
                throw new Error(`@primaryKey configuration of '${klass.namespaceName}.${klass.name}' is invalid; please specify just once "@primaryKey col1,col2"`);
            }
            const { spec: pkSpec, tags, description } = parseConstraintSpec(klass.tags.primaryKey);
            const columns = parseSqlColumnArray(pkSpec);
            const attributes = attributesByNames(klass, columns, `@primaryKey ${klass.tags.primaryKey}`);
            attributes.forEach(attr => {
                attr.tags.notNull = true;
            });
            const keyAttributeNums = attributes.map(a => a.num);
            // Now we need to fake a constraint for this:
            const fakeConstraint = {
                kind: PgEntityKind.CONSTRAINT,
                isFake: true,
                isIndexed: true,
                id: String(Math.random()),
                name: `FAKE_${klass.namespaceName}_${klass.name}_primaryKey`,
                type: "p",
                classId: klass.id,
                foreignClassId: void 0,
                comment: void 0,
                description,
                keyAttributeNums,
                foreignKeyAttributeNums: [],
                tags,
                // Too early for linking
                class: null,
                foreignClass: null,
                keyAttributes: null,
                foreignKeyAttributes: null,
                namespace: null,
            };
            introspectionResults.constraint.push(fakeConstraint);
        }
    });
    // Now primary keys are in place, we can apply foreign keys
    introspectionResults.class.forEach(klass => {
        const namespace = introspectionResults.namespace.find(n => n.id === klass.namespaceId);
        if (!namespace) {
            return;
        }
        const getType = () => introspectionResults.type.find(t => t.id === klass.typeId);
        const foreignKey = klass.tags.foreignKey || getType().tags.foreignKey;
        if (foreignKey) {
            const foreignKeys = typeof foreignKey === "string" ? [foreignKey] : foreignKey;
            if (!Array.isArray(foreignKeys)) {
                throw new Error(`Invalid foreign key smart comment specified on '${klass.namespaceName}.${klass.name}'`);
            }
            foreignKeys.forEach((fkSpecRaw, index) => {
                if (typeof fkSpecRaw !== "string") {
                    throw new Error(`Invalid foreign key spec (${index}) on '${klass.namespaceName}.${klass.name}'`);
                }
                const { spec: fkSpec, tags, description } = parseConstraintSpec(fkSpecRaw);
                const matches = fkSpec.match(/^\(([^()]+)\) references ([^().]+)(?:\.([^().]+))?(?:\s*\(([^()]+)\))?$/i);
                if (!matches) {
                    throw new Error(`Invalid foreignKey syntax for '${klass.namespaceName}.${klass.name}'; expected something like "(col1,col2) references schema.table (c1, c2)", you passed '${fkSpecRaw}'`);
                }
                const [, rawColumns, rawSchemaOrTable, rawTableOnly, rawForeignColumns,] = matches;
                const rawSchema = rawTableOnly
                    ? rawSchemaOrTable
                    : `"${klass.namespaceName}"`;
                const rawTable = rawTableOnly || rawSchemaOrTable;
                const columns = parseSqlColumnArray(rawColumns);
                const foreignSchema = parseSqlColumnString(rawSchema);
                const foreignTable = parseSqlColumnString(rawTable);
                const foreignColumns = rawForeignColumns
                    ? parseSqlColumnArray(rawForeignColumns)
                    : null;
                const foreignKlass = introspectionResults.class.find(k => k.name === foreignTable && k.namespaceName === foreignSchema);
                if (!foreignKlass) {
                    throw new Error(`@foreignKey smart comment referenced non-existant table/view '${foreignSchema}'.'${foreignTable}'. Note that this reference must use *database names* (i.e. it does not respect @name). (${fkSpecRaw})`);
                }
                const foreignNamespace = introspectionResults.namespace.find(n => n.id === foreignKlass.namespaceId);
                if (!foreignNamespace) {
                    return;
                }
                const keyAttributeNums = attributesByNames(klass, columns, `@foreignKey ${fkSpecRaw}`).map(a => a.num);
                const foreignKeyAttributeNums = attributesByNames(foreignKlass, foreignColumns, `@foreignKey ${fkSpecRaw}`).map(a => a.num);
                // Now we need to fake a constraint for this:
                const fakeConstraint = {
                    kind: PgEntityKind.CONSTRAINT,
                    isFake: true,
                    isIndexed: true,
                    id: String(Math.random()),
                    name: `FAKE_${klass.namespaceName}_${klass.name}_foreignKey_${index}`,
                    type: "f",
                    classId: klass.id,
                    foreignClassId: foreignKlass.id,
                    comment: void 0,
                    description,
                    keyAttributeNums,
                    foreignKeyAttributeNums,
                    tags,
                    // Too early for linking
                    class: null,
                    foreignClass: null,
                    keyAttributes: null,
                    foreignKeyAttributes: null,
                    namespace: null,
                };
                introspectionResults.constraint.push(fakeConstraint);
            });
        }
    });
}
exports.default = (async function PgIntrospectionPlugin(builder, { pgConfig, pgSchemas: schemas, pgEnableTags, persistentMemoizeWithKey = (_key, fn) => fn(), pgThrowOnMissingSchema = false, pgIncludeExtensionResources = false, pgLegacyFunctionsOnly = false, pgSkipInstallingWatchFixtures = false, pgAugmentIntrospectionResults, pgOwnerConnectionString, }) {
    const augment = (introspectionResults) => {
        [pgAugmentIntrospectionResults, smartCommentConstraints].forEach(fn => fn ? fn(introspectionResults) : null);
    };
    /**
     * Introspect database and get the table/view/constraints.
     */
    async function introspect() {
        // Perform introspection
        if (!Array.isArray(schemas)) {
            throw new Error("Argument 'schemas' (array) is required");
        }
        const cacheKey = `PgIntrospectionPlugin-introspectionResultsByKind-v${package_json_1.version}`;
        const cloneResults = (obj) => {
            const result = Object.keys(obj).reduce((memo, k) => {
                memo[k] = Array.isArray(obj[k])
                    ? obj[k].map((v) => (Object.assign({}, v)))
                    : obj[k];
                return memo;
            }, {});
            return result;
        };
        const rawishIntrospectionResultsByKind = cloneResults(await persistentMemoizeWithKey(cacheKey, () => withPgClient_1.default(pgConfig, async (pgClient) => {
            const versionResult = await pgClient.query("show server_version_num;");
            const serverVersionNum = parseInt(versionResult.rows[0].server_version_num, 10);
            const introspectionQuery = introspectionQuery_1.makeIntrospectionQuery(serverVersionNum, {
                pgLegacyFunctionsOnly,
            });
            const { rows } = await pgClient.query(introspectionQuery, [
                schemas,
                pgIncludeExtensionResources,
            ]);
            const result = {
                __pgVersion: serverVersionNum,
                namespace: [],
                class: [],
                attribute: [],
                type: [],
                constraint: [],
                procedure: [],
                extension: [],
                index: [],
            };
            for (const { object } of rows) {
                result[object.kind].push(object);
            }
            // Parse tags from comments
            [
                "namespace",
                "class",
                "attribute",
                "type",
                "constraint",
                "procedure",
                "extension",
                "index",
            ].forEach(kind => {
                result[kind].forEach((object) => {
                    // Keep a copy of the raw comment
                    object.comment = object.description;
                    if (pgEnableTags && object.description) {
                        const parsed = utils_1.parseTags(object.description);
                        object.tags = parsed.tags;
                        object.description = parsed.text;
                    }
                    else {
                        object.tags = {};
                    }
                });
            });
            const extensionConfigurationClassIds = flatMap(result.extension, e => e.configurationClassIds);
            result.class.forEach(klass => {
                klass.isExtensionConfigurationTable =
                    extensionConfigurationClassIds.indexOf(klass.id) >= 0;
            });
            [
                "namespace",
                "class",
                "attribute",
                "type",
                "constraint",
                "procedure",
                "extension",
                "index",
            ].forEach(k => {
                result[k].forEach(Object.freeze);
            });
            return Object.freeze(result);
        })));
        const knownSchemas = rawishIntrospectionResultsByKind.namespace.map(n => n.name);
        const missingSchemas = schemas.filter(s => knownSchemas.indexOf(s) < 0);
        if (missingSchemas.length) {
            const errorMessage = `You requested to use schema '${schemas.join("', '")}'; however we couldn't find some of those! Missing schemas are: '${missingSchemas.join("', '")}'`;
            if (pgThrowOnMissingSchema) {
                throw new Error(errorMessage);
            }
            else {
                console.warn("⚠️ WARNING⚠️  " + errorMessage); // eslint-disable-line no-console
            }
        }
        const xByY = (arrayOfX, attrKey) => arrayOfX.reduce((memo, x) => {
            memo[x[attrKey]] = x;
            return memo;
        }, {});
        const xByYAndZ = (arrayOfX, attrKey, attrKey2) => arrayOfX.reduce((memo, x) => {
            if (!memo[x[attrKey]])
                memo[x[attrKey]] = {};
            memo[x[attrKey]][x[attrKey2]] = x;
            return memo;
        }, {});
        const introspectionResultsByKind = Object.assign(Object.assign({}, rawishIntrospectionResultsByKind), { namespaceById: xByY(rawishIntrospectionResultsByKind.namespace, "id"), classById: xByY(rawishIntrospectionResultsByKind.class, "id"), typeById: xByY(rawishIntrospectionResultsByKind.type, "id"), attributeByClassIdAndNum: xByYAndZ(rawishIntrospectionResultsByKind.attribute, "classId", "num"), extensionById: xByY(rawishIntrospectionResultsByKind.extension, "id") });
        function relate(array, newAttr, lookupAttr, lookup, missingOk = false) {
            array.forEach(entry => {
                const key = entry[lookupAttr];
                if (Array.isArray(key)) {
                    entry[newAttr] = key
                        .map(innerKey => {
                        const result = lookup[innerKey];
                        if (innerKey && !result) {
                            if (missingOk) {
                                return;
                            }
                            throw new Error(`Could not look up '${newAttr}' by '${lookupAttr}' ('${innerKey}') on '${JSON.stringify(entry)}'`);
                        }
                        return result;
                    })
                        .filter(_ => _);
                }
                else {
                    const result = lookup[key];
                    if (key && !result) {
                        if (missingOk) {
                            return;
                        }
                        throw new Error(`Could not look up '${newAttr}' by '${lookupAttr}' on '${JSON.stringify(entry)}'`);
                    }
                    entry[newAttr] = result;
                }
            });
        }
        augment(introspectionResultsByKind);
        relate(introspectionResultsByKind.class, "namespace", "namespaceId", introspectionResultsByKind.namespaceById, true // Because it could be a type defined in a different namespace - which is fine so long as we don't allow querying it directly
        );
        relate(introspectionResultsByKind.class, "type", "typeId", introspectionResultsByKind.typeById);
        relate(introspectionResultsByKind.attribute, "class", "classId", introspectionResultsByKind.classById);
        relate(introspectionResultsByKind.attribute, "type", "typeId", introspectionResultsByKind.typeById);
        relate(introspectionResultsByKind.procedure, "namespace", "namespaceId", introspectionResultsByKind.namespaceById);
        relate(introspectionResultsByKind.type, "class", "classId", introspectionResultsByKind.classById, true);
        relate(introspectionResultsByKind.type, "domainBaseType", "domainBaseTypeId", introspectionResultsByKind.typeById, true // Because not all types are domains
        );
        relate(introspectionResultsByKind.type, "arrayItemType", "arrayItemTypeId", introspectionResultsByKind.typeById, true // Because not all types are arrays
        );
        relate(introspectionResultsByKind.constraint, "class", "classId", introspectionResultsByKind.classById);
        relate(introspectionResultsByKind.constraint, "foreignClass", "foreignClassId", introspectionResultsByKind.classById, true // Because many constraints don't apply to foreign classes
        );
        relate(introspectionResultsByKind.extension, "namespace", "namespaceId", introspectionResultsByKind.namespaceById, true // Because the extension could be a defined in a different namespace
        );
        relate(introspectionResultsByKind.extension, "configurationClasses", "configurationClassIds", introspectionResultsByKind.classById, true // Because the configuration table could be a defined in a different namespace
        );
        relate(introspectionResultsByKind.index, "class", "classId", introspectionResultsByKind.classById);
        // Reverse arrayItemType -> arrayType
        introspectionResultsByKind.type.forEach(type => {
            if (type.arrayItemType) {
                type.arrayItemType.arrayType = type;
            }
        });
        // Table/type columns / constraints
        introspectionResultsByKind.class.forEach(klass => {
            klass.attributes = introspectionResultsByKind.attribute.filter(attr => attr.classId === klass.id);
            klass.canUseAsterisk = !klass.attributes.some(attr => attr.columnLevelSelectGrant);
            klass.constraints = introspectionResultsByKind.constraint.filter(constraint => constraint.classId === klass.id);
            klass.foreignConstraints = introspectionResultsByKind.constraint.filter(constraint => constraint.foreignClassId === klass.id);
            klass.primaryKeyConstraint = klass.constraints.find(constraint => constraint.type === "p");
        });
        // Constraint attributes
        introspectionResultsByKind.constraint.forEach(constraint => {
            if (constraint.keyAttributeNums && constraint.class) {
                constraint.keyAttributes = constraint.keyAttributeNums.map(nr => {
                    const attr = constraint.class.attributes.find(attr => attr.num === nr);
                    if (!attr) {
                        throw new Error(`Could not find attribute with index '${nr}' on '${PgBasicsPlugin_1.describePgEntity(constraint.class)}`);
                    }
                    return attr;
                });
            }
            else {
                constraint.keyAttributes = [];
            }
            const { foreignClass } = constraint;
            if (constraint.foreignKeyAttributeNums && foreignClass) {
                constraint.foreignKeyAttributes = constraint.foreignKeyAttributeNums.map(nr => {
                    const attr = foreignClass.attributes.find(attr => attr.num === nr);
                    if (!attr) {
                        throw new Error(`Could not find attribute with index '${nr}' on '${PgBasicsPlugin_1.describePgEntity(constraint.class)}`);
                    }
                    return attr;
                });
            }
            else {
                constraint.foreignKeyAttributes = [];
            }
        });
        // Detect which columns and constraints are indexed
        introspectionResultsByKind.index.forEach(index => {
            const columns = index.attributeNums.map(nr => index.class.attributes.find(attr => attr.num === nr));
            // Indexed column (for orderBy / filter):
            if (columns[0]) {
                columns[0].isIndexed = true;
            }
            if (columns[0] && columns.length === 1 && index.isUnique) {
                columns[0].isUnique = true;
            }
            // Indexed constraints (for reverse relations):
            index.class.constraints
                .filter(constraint => constraint.type === "f")
                .forEach(constraint => {
                if (constraint.keyAttributeNums.every((nr, idx) => index.attributeNums[idx] === nr)) {
                    constraint.isIndexed = true;
                }
            });
        });
        return introspectionResultsByKind;
    }
    let introspectionResultsByKind = await introspect();
    let listener = null;
    class Listener {
        constructor(triggerRebuild) {
            this.stopped = false;
            this._handleChange = throttle(async () => {
                debug(`Schema change detected: re-inspecting schema...`);
                try {
                    introspectionResultsByKind = await introspect();
                    debug(`Schema change detected: re-inspecting schema complete`);
                    triggerRebuild();
                }
                catch (e) {
                    // eslint-disable-next-line no-console
                    console.error(`Schema introspection failed: ${e.message}`);
                }
            }, 750, {
                leading: true,
                trailing: true,
            });
            this._listener = this._listener.bind(this);
            this._handleClientError = this._handleClientError.bind(this);
            this._start();
        }
        async _start(isReconnect = false) {
            if (this.stopped) {
                return;
            }
            // Connect to DB
            try {
                const { pgClient, releasePgClient, } = await withPgClient_1.getPgClientAndReleaserFromConfig(pgConfig);
                this.client = pgClient;
                // $FlowFixMe: hack property
                this._reallyReleaseClient = releasePgClient;
                pgClient.on("notification", this._listener);
                pgClient.on("error", this._handleClientError);
                if (this.stopped) {
                    // In case watch mode was cancelled in the interrim.
                    return this._releaseClient();
                }
                else {
                    await pgClient.query("listen postgraphile_watch");
                    // Install the watch fixtures.
                    if (!pgSkipInstallingWatchFixtures) {
                        const watchSqlInner = await readFile(WATCH_FIXTURES_PATH, "utf8");
                        const sql = `begin; ${watchSqlInner};`;
                        await withPgClient_1.default(pgOwnerConnectionString || pgConfig, async (pgClient) => {
                            try {
                                await pgClient.query(sql);
                            }
                            catch (error) {
                                if (!this._haveDisplayedError) {
                                    this._haveDisplayedError = true;
                                    /* eslint-disable no-console */
                                    console.warn(`${chalk_1.default.bold.yellow("Failed to setup watch fixtures in Postgres database")} ️️⚠️`);
                                    console.warn(chalk_1.default.yellow("This is likely because the PostgreSQL user in the connection string does not have sufficient privileges; you can solve this by passing the 'owner' connection string via '--owner-connection' / 'ownerConnectionString'. If the fixtures already exist, the watch functionality may still work."));
                                    console.warn(chalk_1.default.yellow("Enable DEBUG='graphile-build-pg' to see the error"));
                                    /* eslint-enable no-console */
                                }
                                debug(error);
                            }
                            finally {
                                await pgClient.query("commit;");
                            }
                        });
                    }
                    // Trigger re-introspection on server reconnect
                    if (isReconnect) {
                        this._handleChange();
                    }
                }
            }
            catch (e) {
                // If something goes wrong, disconnect and try again after a short delay
                this._reconnect(e);
            }
        }
        _handleClientError(e) {
            // Client is already cleaned up
            this.client = null;
            this._reallyReleaseClient = null;
            this._reconnect(e);
        }
        async _reconnect(e) {
            if (this.stopped) {
                return;
            }
            // eslint-disable-next-line no-console
            console.error("Error occurred for PG watching client; reconnecting in 2 seconds.", e.message);
            await this._releaseClient();
            setTimeout(() => {
                if (!this.stopped) {
                    // Listen for further changes
                    this._start(true);
                }
            }, 2000);
        }
        async _listener(notification) {
            if (notification.channel !== "postgraphile_watch") {
                return;
            }
            try {
                // See watch-fixtures.sql for types
                const payload = JSON.parse(notification.payload);
                if (payload.type === "ddl") {
                    const commands = (payload.payload || [])
                        .filter(({ schema }) => schema == null || schemas.indexOf(schema) >= 0)
                        .map(({ command }) => command);
                    if (commands.length) {
                        this._handleChange();
                    }
                }
                else if (payload.type === "drop") {
                    const affectsOurSchemas = (payload.payload || []).some(schemaName => schemas.indexOf(schemaName) >= 0);
                    if (affectsOurSchemas) {
                        this._handleChange();
                    }
                }
                else {
                    throw new Error(`Payload type '${payload.type}' not recognised`);
                }
            }
            catch (e) {
                debug(`Error occurred parsing notification payload: ${e}`);
            }
        }
        async stop() {
            this.stopped = true;
            await this._releaseClient();
        }
        async _releaseClient() {
            // $FlowFixMe
            this._handleChange.cancel();
            const pgClient = this.client;
            const reallyReleaseClient = this._reallyReleaseClient;
            this.client = null;
            this._reallyReleaseClient = null;
            if (pgClient) {
                pgClient.query("unlisten postgraphile_watch").catch(e => {
                    debug(`Error occurred trying to unlisten watch: ${e}`);
                });
                pgClient.removeListener("notification", this._listener);
                pgClient.removeListener("error", this._handleClientError);
                if (reallyReleaseClient) {
                    await reallyReleaseClient();
                }
            }
        }
    }
    builder.registerWatcher(async (triggerRebuild) => {
        // In case we started listening before, clean up
        if (listener) {
            await listener.stop();
        }
        // We're not worried about a race condition here.
        // eslint-disable-next-line require-atomic-updates
        listener = new Listener(triggerRebuild);
    }, async () => {
        const l = listener;
        listener = null;
        if (l) {
            await l.stop();
        }
    });
    builder.hook("build", build => {
        if (introspectionResultsByKind.__pgVersion < 90500) {
            // TODO:v5: remove this workaround
            // This is a bit of a hack, but until we have plugin priorities it's the
            // easiest way to conditionally support PG9.4.
            build.pgQueryFromResolveData = queryFromResolveDataFactory_1.default({
                supportsJSONB: false,
            });
        }
        return build.extend(build, {
            pgIntrospectionResultsByKind: introspectionResultsByKind,
        }, "Adding PG introspection results from graphile-build-pg PgIntrospectionPlugin");
    }, ["PgIntrospection"], [], ["PgBasics"]);
});
//# sourceMappingURL=PgIntrospectionPlugin.js.map