"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const pg = require("pg");
const debug_1 = require("debug");
const debug = debug_1.default("graphile-build-pg");
function constructorName(obj) {
    return ((typeof obj === "object" &&
        obj &&
        typeof obj.constructor === "function" &&
        obj.constructor.name) ||
        null);
}
// Some duck-typing
function quacksLikePgClient(pgConfig) {
    // A diagnosis of exclusion
    if (typeof pgConfig !== "object" || !pgConfig)
        return false;
    if (constructorName(pgConfig) !== "Client")
        return false;
    if (typeof pgConfig["connect"] !== "function")
        return false;
    if (typeof pgConfig["end"] !== "function")
        return false;
    if (typeof pgConfig["escapeLiteral"] !== "function")
        return false;
    if (typeof pgConfig["escapeIdentifier"] !== "function")
        return false;
    return true;
}
function quacksLikePgPool(pgConfig) {
    // A diagnosis of exclusion
    if (typeof pgConfig !== "object" || !pgConfig)
        return false;
    if (constructorName(pgConfig) !== "Pool" &&
        constructorName(pgConfig) !== "BoundPool") {
        return false;
    }
    if (!pgConfig["Client"])
        return false;
    if (!pgConfig["options"])
        return false;
    if (typeof pgConfig["connect"] !== "function")
        return false;
    if (typeof pgConfig["end"] !== "function")
        return false;
    if (typeof pgConfig["query"] !== "function")
        return false;
    return true;
}
exports.quacksLikePgPool = quacksLikePgPool;
exports.getPgClientAndReleaserFromConfig = async (pgConfig = process.env
    .DATABASE_URL) => {
    let releasePgClient = () => { };
    let pgClient;
    if (pgConfig instanceof pg.Client || quacksLikePgClient(pgConfig)) {
        pgClient = pgConfig;
        if (!pgClient.release) {
            throw new Error("We only support PG clients from a PG pool (because otherwise the `await` call can hang indefinitely if an error occurs and there's no error handler)");
        }
    }
    else if (pgConfig instanceof pg.Pool || quacksLikePgPool(pgConfig)) {
        const pgPool = pgConfig;
        const client = await pgPool.connect();
        pgClient = client;
        releasePgClient = () => client.release();
    }
    else if (pgConfig === undefined || typeof pgConfig === "string") {
        const client = new pg.Client(pgConfig);
        pgClient = client;
        client.on("error", e => {
            debug("pgClient error occurred: %s", e);
        });
        releasePgClient = () => new Promise((resolve, reject) => client.end(err => (err ? reject(err) : resolve())));
        await new Promise((resolve, reject) => client.connect(err => (err ? reject(err) : resolve())));
    }
    else {
        throw new Error("You must provide either a pg.Pool or pg.Client instance or a PostgreSQL connection string.");
    }
    return { pgClient, releasePgClient };
};
const withPgClient = async (pgConfig, fn) => {
    if (!fn) {
        throw new Error("Nothing to do!");
    }
    const { pgClient, releasePgClient } = await exports.getPgClientAndReleaserFromConfig(pgConfig);
    const errorHandler = (e) => {
        // eslint-disable-next-line no-console
        console.error("withPgClient client error:", e.message);
    };
    pgClient.on("error", errorHandler);
    try {
        return await fn(pgClient);
    }
    finally {
        pgClient.removeListener("error", errorHandler);
        try {
            await releasePgClient();
        }
        catch (e) {
            // Failed to release, assuming success
        }
    }
};
exports.default = withPgClient;
//# sourceMappingURL=withPgClient.js.map